
/* Hi Pavel,

   this was changed by me:
      * printf -> printk
      * use MALLOC instead of stack in sq_dec (otherwise serious crashes)
      
*/

#include <linux/sched.h>
#include <linux/ctype.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <asm/segment.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/msdos_fs.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/shm.h>
#include <linux/mman.h>
#include <asm/system.h>
#include <asm/byteorder.h>
#include "dmsdos.h"

#if !defined(le16_to_cpu)
 /* for old kernel versions - works only on i386 */
 #define le16_to_cpu(v) (v)
#endif

typedef long int pos_t;
typedef long int sect_t;

#ifdef __GNUC__ 
 #ifdef __i386__
  #define USE_GNU_ASM_i386
 #endif
#endif

/* high speed compare and move routines */
#ifdef USE_GNU_ASM_i386

/* copy block, overlaping part is replaced by repeat of previous part */
/* pointers and counter are modified to point after block */
#define M_MOVSB(D,S,C) \
__asm__ /*__volatile__*/(\
	"cld\n\t" \
	"rep\n\t" \
	"movsb\n" \
	:"=D" (D),"=S" (S),"=c" (C) \
	:"0" (D),"1" (S),"2" (C) \
	:"memory")

/* compare blocks, overlaping repeat test */
/* pointers and counter are modified to point after block */
/* D and S points to first diff adr */
#define M_FIRSTDIFF(D,S,C) \
__asm__ /*__volatile__*/(\
	"cld\n\t" \
	"rep\n\t" \
	"cmpsb\n\t" \
	"je 1f\n\t" \
	"dec %0\n\t" \
	"dec %1\n\t" \
	"1:\n" \
	:"=D" (D),"=S" (S),"=c" (C) \
	:"0" (D),"1" (S),"2" (C) \
	)

#else

#define M_MOVSB(D,S,C) for(;(C);(C)--) *((__u8*)(D)++)=*((__u8*)(S)++)
#define M_FIRSTDIFF(D,S,C) for(;(*(__u8*)(D)==*(__u8*)(S))&&(C);\
                           (__u8*)(D)++,(__u8*)(S)++,(C)--)

#endif

#define	INLINE	inline

/*==============================================================*/
/* bitstream reading */

/* for reading and writting from/to bitstream */
typedef
 struct {
   __u32 buf;	/* bit buffer */
     int pb;	/* already readed bits from buf */
   __u16 *pd;	/* first not readed input data */
   __u16 *pe;	/* after end of data */
 } bits_t;

const unsigned sq_bmsk[]=
   {0x0,0x1,0x3,0x7,0xF,0x1F,0x3F,0x7F,0xFF,
    0x1FF,0x3FF,0x7FF,0xFFF,0x1FFF,0x3FFF,0x7FFF,0xFFFF};

/* read next 16 bits from input */
#define RDN_G16(bits) \
   { \
    (bits).buf>>=16; \
    (bits).pb-=16; \
    if((bits).pd<(bits).pe) \
    { \
     (bits).buf|=((__u32)(le16_to_cpu(*((bits).pd++))))<<16; \
    }; \
   }

/* prepares at least 16 bits for reading */
#define RDN_PR(bits,u) \
   { \
    if((bits).pb>=16) RDN_G16(bits); \
    u=(bits).buf>>(bits).pb; \
   }

/* initializes reading from bitstream */
INLINE void sq_rdi(bits_t *pbits,void *pin,unsigned lin)
{
  pbits->pb=32;
  pbits->pd=(__u16*)pin;
  pbits->pe=pbits->pd+((lin+1)>>1);
};

/* reads n<=16 bits from bitstream *pbits */
INLINE unsigned sq_rdn(bits_t *pbits,int n)
{
  unsigned u;
  RDN_PR(*pbits,u);
  pbits->pb+=n;
  u&=sq_bmsk[n];
  return u;
};

/*==============================================================*/
/* huffman decoding */

#define MAX_SPDA_BITS	10
#define MAX_SPDA_LEN	(1<<MAX_SPDA_BITS)
#define MAX_BITS	16
#define MAX_CODES	0x200
#define OUT_OVER	0x100

typedef
 struct {
  __s8 ln;			/* character code */
  __u8 ch;			/* character lens .. if >=0x80 token */
 }huf_chln_t;

typedef
 struct {
  huf_chln_t chln[MAX_SPDA_LEN];/* character codes decode array */
  huf_chln_t chln1[MAX_CODES];	/* for codes with more than bn bits */
  int  bn;			/* ch,ln arrays convert max bn bits codes */
  unsigned cd_ln[MAX_BITS+1];	/* distribution of bits */
  unsigned cd_ch[MAX_BITS+1];	/* distribution of codes codes */
 }huf_rd_t;

__u8 swap_bits_xlat[]=
	  {0x00,0x80,0x40,0xc0,0x20,0xa0,0x60,0xe0,0x10,0x90,
	   0x50,0xd0,0x30,0xb0,0x70,0xf0,0x08,0x88,0x48,0xc8,
	   0x28,0xa8,0x68,0xe8,0x18,0x98,0x58,0xd8,0x38,0xb8,
	   0x78,0xf8,0x04,0x84,0x44,0xc4,0x24,0xa4,0x64,0xe4,
	   0x14,0x94,0x54,0xd4,0x34,0xb4,0x74,0xf4,0x0c,0x8c,
	   0x4c,0xcc,0x2c,0xac,0x6c,0xec,0x1c,0x9c,0x5c,0xdc,
	   0x3c,0xbc,0x7c,0xfc,0x02,0x82,0x42,0xc2,0x22,0xa2,
	   0x62,0xe2,0x12,0x92,0x52,0xd2,0x32,0xb2,0x72,0xf2,
	   0x0a,0x8a,0x4a,0xca,0x2a,0xaa,0x6a,0xea,0x1a,0x9a,
	   0x5a,0xda,0x3a,0xba,0x7a,0xfa,0x06,0x86,0x46,0xc6,
	   0x26,0xa6,0x66,0xe6,0x16,0x96,0x56,0xd6,0x36,0xb6,
	   0x76,0xf6,0x0e,0x8e,0x4e,0xce,0x2e,0xae,0x6e,0xee,
	   0x1e,0x9e,0x5e,0xde,0x3e,0xbe,0x7e,0xfe,0x01,0x81,
	   0x41,0xc1,0x21,0xa1,0x61,0xe1,0x11,0x91,0x51,0xd1,
	   0x31,0xb1,0x71,0xf1,0x09,0x89,0x49,0xc9,0x29,0xa9,
	   0x69,0xe9,0x19,0x99,0x59,0xd9,0x39,0xb9,0x79,0xf9,
	   0x05,0x85,0x45,0xc5,0x25,0xa5,0x65,0xe5,0x15,0x95,
	   0x55,0xd5,0x35,0xb5,0x75,0xf5,0x0d,0x8d,0x4d,0xcd,
	   0x2d,0xad,0x6d,0xed,0x1d,0x9d,0x5d,0xdd,0x3d,0xbd,
	   0x7d,0xfd,0x03,0x83,0x43,0xc3,0x23,0xa3,0x63,0xe3,
	   0x13,0x93,0x53,0xd3,0x33,0xb3,0x73,0xf3,0x0b,0x8b,
	   0x4b,0xcb,0x2b,0xab,0x6b,0xeb,0x1b,0x9b,0x5b,0xdb,
	   0x3b,0xbb,0x7b,0xfb,0x07,0x87,0x47,0xc7,0x27,0xa7,
	   0x67,0xe7,0x17,0x97,0x57,0xd7,0x37,0xb7,0x77,0xf7,
	   0x0f,0x8f,0x4f,0xcf,0x2f,0xaf,0x6f,0xef,0x1f,0x9f,
	   0x5f,0xdf,0x3f,0xbf,0x7f,0xff};

/* swap 16 bits order */
INLINE unsigned swap_bits_order_16(unsigned d)
{ unsigned r;
  #ifdef USE_GNU_ASM_i386
    __asm__ (
	"xlatl\n\t"
	"xchgb	%%al,%%ah\n\t"
	"xlatl\n\t"
	:"=a"(r):"0"(d),"b"(swap_bits_xlat)); 
  #else
    r=((unsigned)swap_bits_xlat[(__u8)d])<<8;
    r|=swap_bits_xlat[(__u8)(d>>8)];
  #endif
  return r;
};

/* swap bit order */
INLINE unsigned swap_bits_order(unsigned d,int n)
{ unsigned r=0;
  while(n--) { r<<=1;r|=d&1;d>>=1;};
  return r;
};

/* initializes huffman conversion structure *phuf for m codes,
   *ca code and token bit lengths, ends with 0xFF, 
   bn predicated maximal bit length */
int sq_rdhufi(huf_rd_t *phuf,int m,int bn,__u8 *ca)
{
 if(bn>MAX_SPDA_BITS) bn=MAX_SPDA_BITS;
 phuf->bn=bn;
 {
  int i;
  unsigned u,us,ut;
  memset(phuf->cd_ln,0,sizeof(phuf->cd_ln));i=0;
  while((u=ca[i++])<=MAX_BITS) phuf->cd_ln[u]++;
  memset(phuf->cd_ch,0,sizeof(phuf->cd_ch));
  phuf->cd_ln[0]=0;us=0;ut=0;
  for(i=1;i<=MAX_BITS;i++)
  {
   u=phuf->cd_ln[i];phuf->cd_ln[i]=ut;
   phuf->cd_ch[i]=us;ut+=u;us+=u;us<<=1;
  };
  /* if suceed us should be 0x10000 */ 
  if (us&((1<<MAX_BITS)-1)) return(0);
 };
 {
  int i,ln,l,ch,sh,cod;
  for(i=0;(l=ln=ca[i])<=MAX_BITS;i++) if(ln)
  {
   sh=(bn-ln);
   cod=(phuf->cd_ch[ln])++;
   cod=swap_bits_order_16(cod)>>(16-ln);
   if(i<m) ch=i; else {ch=i-m+1;ln-=0x40;};
   if (sh>0)
   {
    sh=1<<sh;
    l=1<<l;
    while(sh--)
    {
      phuf->chln[cod].ch=ch;
      phuf->chln[cod].ln=ln;
      cod+=l;
    };
   } else if (sh==0) {
    phuf->chln[cod].ch=ch;
    phuf->chln[cod].ln=ln;
   } else {
    cod&=sq_bmsk[bn];
    phuf->chln[cod].ch=0x00;
    phuf->chln[cod].ln=-0x40;
    cod=(phuf->cd_ln[l])++;
    phuf->chln1[cod].ch=ch;
    phuf->chln1[cod].ln=ln;
   };
  };
  /* if suceed ln should be 0xFF */
 };
 return(1);
};

/* read and huffman decode of characters, stops on tokens or buffer ends */
INLINE unsigned sq_rdh(bits_t *pbits,const huf_rd_t *phuf,__u8 **pout,__u8 *pend)
{
 unsigned ch;
 unsigned bmsk=sq_bmsk[phuf->bn];

 while(1)
 {while(1)
  {if(pbits->pb>=16)
    RDN_G16(*pbits);
   if (*pout>=pend) return OUT_OVER;
   ch=(pbits->buf>>pbits->pb)&bmsk;
   if((pbits->pb+=phuf->chln[ch].ln)<0) break;
   *((*pout)++)=phuf->chln[ch].ch;

   if(pbits->pb<16)
   {if (*pout>=pend) return OUT_OVER;
    ch=(pbits->buf>>pbits->pb)&bmsk;
    if((pbits->pb+=phuf->chln[ch].ln)<0) break;
    *((*pout)++)=phuf->chln[ch].ch;

    if(pbits->pb<16)
    {if (*pout>=pend) return OUT_OVER;
     ch=(pbits->buf>>pbits->pb)&bmsk;
     if((pbits->pb+=phuf->chln[ch].ln)<0) break;
     *((*pout)++)=phuf->chln[ch].ch;
    };
   };
  };

  ch=phuf->chln[ch].ch;
  pbits->pb+=0x40; if(ch--) return ch;
  /* code longer than phuf->bn */
  if(pbits->pb>=16) RDN_G16(*pbits);
  ch=swap_bits_order_16((__u16)(pbits->buf>>pbits->pb));
  {
   int i;
   i=phuf->bn;
   do
    i++;
   while((phuf->cd_ch[i]<=(ch>>(16-i)))&&(i<MAX_BITS));
   ch=((ch>>(16-i)))-phuf->cd_ch[i]+phuf->cd_ln[i];
  };
  if((pbits->pb+=phuf->chln1[ch].ln)<0) 
  {pbits->pb+=0x40;
   return phuf->chln1[ch].ch-1;
  };
  *((*pout)++)=phuf->chln1[ch].ch;
 };
};

/* read one huffman encoded value */
INLINE unsigned sq_rdh1(bits_t *pbits,const huf_rd_t *phuf)
{unsigned ch;
 if(pbits->pb>=16) RDN_G16(*pbits);
 ch=(pbits->buf>>pbits->pb)&sq_bmsk[phuf->bn];
 if((pbits->pb+=phuf->chln[ch].ln)>=0) return phuf->chln[ch].ch;
 ch=phuf->chln[ch].ch;
 pbits->pb+=0x40; if(ch) return ch+0x100-1;
 ch=swap_bits_order_16((__u16)(pbits->buf>>pbits->pb));
 {int i;
  i=phuf->bn;
  do
   i++;
  while((phuf->cd_ch[i]<=(ch>>(16-i)))&&(i<MAX_BITS));
  ch=((ch>>(16-i)))-phuf->cd_ch[i]+phuf->cd_ln[i];
 };
 if((pbits->pb+=phuf->chln1[ch].ln)>=0) return phuf->chln1[ch].ch;
 pbits->pb+=0x40;
 return phuf->chln1[ch].ch+0x100-1;
};

/*==============================================================*/
/* SQ decompression */

/* index conversion table for first bitlen table */
const int code_index_1[]={0x10,0x11,0x12,0x00,0x08,0x07,0x09,0x06,0x0A,0x05,
			  0x0B,0x04,0x0C,0x03,0x0D,0x02,0x0E,0x01,0x0F};

__u16 sqt_repbas[]={0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0D,
		    0x0F,0x11,0x13,0x17,0x1B,0x1F,0x23,0x2B,0x33,0x3B,
		    0x43,0x53,0x63,0x73,0x83,0xA3,0xC3,0xE3,0x102,0x00,0x00};

__u16 sqt_repbln[]={0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x01,
		    0x01,0x01,0x02,0x02,0x02,0x02,0x03,0x03,0x03,0x03,
		    0x04,0x04,0x04,0x04,0x05,0x05,0x05,0x05,0x00,0x63,0x63};

__u16 sqt_offbas[]={0x0001,0x0002,0x0003,0x0004,0x0005,0x0007,0x0009,0x000D,
		    0x0011,0x0019,0x0021,0x0031,0x0041,0x0061,0x0081,0x00C1,
		    0x0101,0x0181,0x0201,0x0301,0x0401,0x0601,0x0801,0x0C01,
		    0x1001,0x1801,0x2001,0x3001,0x4001,0x6001};

__u16 sqt_offbln[]={0x0000,0x0000,0x0000,0x0000,0x0001,0x0001,0x0002,0x0002,
		    0x0003,0x0003,0x0004,0x0004,0x0005,0x0005,0x0006,0x0006,
		    0x0007,0x0007,0x0008,0x0008,0x0009,0x0009,0x000A,0x000A,
		    0x000B,0x000B,0x000C,0x000C,0x000D,0x000D};

int sq_dec(void* pin,int lin, void* pout, int lout, int flg)
{ 
  __u8 *p, *pend, *r;
  unsigned u, u1, repoffs, replen;
  int i,bn_max;
  bits_t bits;
  int final_flag;
  int count_1;
  int count_2;
  int count_3;
  int method;
  unsigned mask;
  __u8 *code_bln=NULL; /*[1024]; ...too large for kernel stack...*/
  /*huf_rd_t huf1,huf2; ...also too large... */
  huf_rd_t*huf1p,*huf2p;
  
  sq_rdi(&bits,pin,lin);
  p=(__u8*)pout;pend=p+lout;
  if((sq_rdn(&bits,8)!='S')||(sq_rdn(&bits,8)!='Q'))
  { printk("DMSDOS: Data are not SQ compressed\n");
    return(0);
  };
  u=sq_rdn(&bits,16);
  LOG_DECOMP("DMSDOS: sq_dec: version %X\n",u);
  
  code_bln=(__u8*)MALLOC(1024*sizeof(__u8));
  if(code_bln==NULL)
  { printk("DMSDOS: sq_dec: out of memory!\n");
    return -1;
  }
  huf1p=(huf_rd_t*)MALLOC(sizeof(huf_rd_t));
  if(huf1p==NULL)
  { printk("DMSDOS: sq_dec: out of memory!\n");
    FREE(code_bln);
    return -1;
  }
  huf2p=(huf_rd_t*)MALLOC(sizeof(huf_rd_t));
  if(huf1p==NULL)
  { printk("DMSDOS: sq_dec: out of memory!\n");
    FREE(code_bln);
    FREE(huf1p);
    return -1;
  }
  
  do
  { final_flag=sq_rdn(&bits,1); LOG_DECOMP("DMSDOS: final_flag %d\n",final_flag);
    method=sq_rdn(&bits,2);     LOG_DECOMP("DMSDOS: method %d\n",method);
    switch(method)
    {
      case 0: 
        printk("DMSDOS: dec_sq: submethod not tested - raw read\n");
	/* go to byte boundary */
	/* read 16 bits - count of raw bytes */
	sq_rdn(&bits,(8-bits.pb)&7);
	replen=sq_rdn(&bits,16);
	r=(__u8*)bits.pd-(32-bits.pb)/8;
        if(r+replen>(__u8*)bits.pe) 
        { FREE(code_bln);
          FREE(huf1p);
          FREE(huf2p);
          return 0;
        }
        if(p+replen>pend)
        { FREE(code_bln);
          FREE(huf1p);
          FREE(huf2p);
          return 0;
        }
	M_MOVSB(p,r,replen); /* copy/repeat function */
	if((unsigned)r&1) bits.pb=32;
	else {bits.pb=32+8;r--;};
	bits.pd=(typeof(bits.pd))r;
        break;

      case 1: 
        printk("DMSDOS: sq_dec: submethod not tested - fixed huffman\n");
        /* 0x90*8 0x70*9 0x18*7 8*8 */
        /* 0x1E*5 */
        bn_max=9;
        count_1=0x120;
        count_2=0x1E;
        r=code_bln;
        for(i=0;i<0x90;i++) *(r++)=8;
        for(i=0;i<0x70;i++) *(r++)=9;
        for(i=0;i<0x18;i++) *(r++)=7;
        for(i=0;i<0x08;i++) *(r++)=8;
        for(i=0;i<0x1E;i++) *(r++)=5;
        goto case_1_cont;
      
      case 2:
        LOG_DECOMP("DMSDOS: sq_dec: submethod huffman\n"); /* SQ_dec_sub_C */
        count_1=sq_rdn(&bits,5)+0x101;		/* esp+20 */
        LOG_DECOMP("DMSDOS: count_1 %d\n",count_1);
        if(count_1>0x11E) 
        { printk("DMSDOS: sq_dec: huff count_1 too big\n");
          FREE(code_bln);
          FREE(huf1p);
          FREE(huf2p);
          return(0);
        };
        count_2=sq_rdn(&bits,5)+1;		/* esp+24 */
        LOG_DECOMP("DMSDOS: count_2 %d\n",count_2);
        if(count_2>0x1E) 
        { printk("DMSDOS: sq_dec: huff count_2 too big\n");
          FREE(code_bln);
          FREE(huf1p);
          FREE(huf2p);
          return(0);
        };
        count_3=sq_rdn(&bits,4)+4;		/* esp+14 */
        LOG_DECOMP("DMSDOS: count_3 %d\n",count_3);
        bn_max=0;
        for(i=0;i<count_3;i++)
        { u=sq_rdn(&bits,3);
          code_bln[code_index_1[i]]=u;
          if(u>bn_max)bn_max=u;
        };
        while(i<19) code_bln[code_index_1[i++]]=0;code_bln[19]=0xFF;
        i=sq_rdhufi(huf1p,19,bn_max,code_bln);
        if(!i)
        { printk("DMSDOS: sq_dec: huff error in helper table\n");
          FREE(code_bln);
          FREE(huf1p);FREE(huf2p);
          return 0;
        };
        mask=sq_bmsk[huf1p->bn]; u1=0; bn_max=0;
        for(i=0;i<count_1+count_2;)
        { RDN_PR(bits,u);
          bits.pb+=huf1p->chln[u&mask].ln;
          u=huf1p->chln[u&mask].ch;
          switch(u)
          { case 16:		/* 3 to 6 repeats of last */
              u=sq_rdn(&bits,2)+3;
              while(u--) code_bln[i++]=u1;
              break;
            case 17:		/* 3 to 10 repeats of 0 */
              u=sq_rdn(&bits,3)+3; u1=0;
              while(u--) code_bln[i++]=u1;
              break;
            case 18:		/* 11 to 139 repeats of 0 */
              u=sq_rdn(&bits,7)+11; u1=0;
              while(u--) code_bln[i++]=u1;
              break;
            default:
              code_bln[i++]=u;
              u1=u;
              if(u>bn_max) bn_max=u;
          };
        };

      case_1_cont:
        /* code_bln+count_1	0x96  count_2 sqt_offbln sqt_offbas */
        code_bln[count_1+count_2]=0xFF;
        i=sq_rdhufi(huf2p,0x100,8,code_bln+count_1);
        if(!i) 
        { printk("DMSDOS: sq_dec: huff error in offset table\n");
          FREE(code_bln);FREE(huf1p);FREE(huf2p);
          return 0;
        };
      
        /* code_bln		0x100 count_1 sqt_repbln sqt_repbas */
        code_bln[count_1]=0xFF;
        i=sq_rdhufi(huf1p,0x100,bn_max,code_bln);
        if(!i)
        { printk("DMSDOS: sq_dec: huff error in char and len table\n");
          FREE(code_bln);FREE(huf1p);FREE(huf2p);
          return 0;
        }
      
        while((u=sq_rdh(&bits,huf1p,&p,pend))!=0)
        { if(u==OUT_OVER){u=sq_rdh1(&bits,huf1p)-0x100;break;};
          u--;
          replen=sqt_repbas[u]+sq_rdn(&bits,sqt_repbln[u]);
          u=sq_rdh1(&bits,huf2p);
          repoffs=sqt_offbas[u]+sq_rdn(&bits,sqt_offbln[u]);
          if ((__u8*)pout+repoffs>p) 
          { repoffs=p-(__u8*)pout;
            printk("DMSDOS: sq_dec: huff offset UNDER\n");
          };
          if (p+replen>pend)   
          { replen=pend-p;
            printk("DMSDOS: sq_dec: huff offset OVER\n");
          };
          r=p-repoffs; M_MOVSB(p,r,replen); /* copy/repeat function */
        };
      
        if(u) 
        { printk("DMSDOS: sq_dec: huff BAD last token %x\n",u);
          FREE(code_bln);FREE(huf1p);FREE(huf2p);
          return 0;
        };
        break;
      
      case 3:
        printk("DMSDOS: sq_dec: unknown submethod - 3\n");
        FREE(code_bln);
        FREE(huf1p);
        FREE(huf2p);
        return(0);      
    };
  } while((!final_flag)&&(p<pend));
  
  FREE(code_bln);
  FREE(huf1p);
  FREE(huf2p);
  return(p-(__u8*)pout);
};
